@@ -105,7 +105,7 @@ typedef struct CSVParser
 	bool	   *fnn;			/**< array of NOT NULL column flag */
 } CSVParser;
 
-static void	CSVParserInit(CSVParser *self, Checker *checker, const char *infile, TupleDesc desc);
+static void	CSVParserInit(CSVParser *self, Checker *checker, const char *infile, TupleDesc desc, bool multi_process);
 static HeapTuple	CSVParserRead(CSVParser *self, Checker *checker);
 static int64	CSVParserTerm(CSVParser *self);
 static bool CSVParserParam(CSVParser *self, const char *keyword, char *value);
@@ -178,8 +178,10 @@ CreateCSVParser(void)
  * @note Caller must release the resource using CSVParserTerm().
  */
 static void
-CSVParserInit(CSVParser *self, Checker *checker, const char *infile, TupleDesc desc)
+CSVParserInit(CSVParser *self, Checker *checker, const char *infile, TupleDesc desc, bool multi_process)
 {
+	TupleCheckStatus	status;
+
 	/*
 	 * set default values
 	 */
@@ -208,9 +210,12 @@ CSVParserInit(CSVParser *self, Checker *checker, const char *infile, TupleDesc d
 				 errmsg
 				 ("cannot use FILTER with FORCE_NOT_NULL")));
 
-	self->source = CreateSource(infile, desc);
+	self->source = CreateSource(infile, desc, multi_process);
+
+	status = FilterInit(&self->filter, desc);
+	if (checker->tchecker)
+		checker->tchecker->status = status;
 
-	FilterInit(&self->filter, desc);
 	TupleFormerInit(&self->former, &self->filter, desc);
 
 	/*
