@@ -101,16 +101,18 @@ void
 SpoolerOpen(Spooler *self,
 			Relation rel,
 			bool use_wal,
-			const WriterOptions *options)
+			ON_DUPLICATE on_duplicate,
+			int64 max_dup_errors,
+			const char *dup_badfile)
 {
 	memset(self, 0, sizeof(Spooler));
 
-	self->on_duplicate = options->on_duplicate;
+	self->on_duplicate = on_duplicate;
 	self->use_wal = use_wal;
-	self->max_dup_errors = options->max_dup_errors;
+	self->max_dup_errors = max_dup_errors;
 	self->dup_old = 0;
 	self->dup_new = 0;
-	self->dup_badfile = pstrdup(options->dup_badfile);
+	self->dup_badfile = pstrdup(dup_badfile);
 	self->dup_fp = NULL;
 
 	self->relinfo = makeNode(ResultRelInfo);
@@ -129,7 +131,7 @@ SpoolerOpen(Spooler *self,
 	self->slot = MakeSingleTupleTableSlot(RelationGetDescr(rel));
 
 	self->spools = IndexSpoolBegin(self->relinfo,
-								   options->max_dup_errors == 0);
+								   max_dup_errors == 0);
 }
 
 void
@@ -350,7 +352,7 @@ _bt_mergebuild(Spooler *self, BTSpool *btspool)
 	 * it's not a temp index.
 	 */
 	wstate.btws_use_wal = self->use_wal &&
-		XLogArchivingActive() && !wstate.index->rd_istemp;
+		XLogArchivingActive() && !RELATION_IS_LOCAL(wstate.index);
 
 	/* reserve the metapage */
 	wstate.btws_pages_alloced = BTREE_METAPAGE + 1;
@@ -579,7 +581,7 @@ _bt_mergeload(Spooler *self, BTWriteState *wstate, BTSpool *btspool, BTReader *b
 	 * fsync those pages here, they might still not be on disk when the crash
 	 * occurs.
 	 */
-	if (!wstate->index->rd_istemp)
+	if (!RELATION_IS_LOCAL(wstate->index))
 	{
 		RelationOpenSmgr(wstate->index);
 		smgrimmedsync(wstate->index->rd_smgr, MAIN_FORKNUM);
@@ -634,7 +636,7 @@ BTReaderInit(BTReader *reader, Relation rel)
 #if PG_VERSION_NUM >= 90100
 	reader->smgr.smgr_rnode.node = rel->rd_node;
 	reader->smgr.smgr_rnode.backend =
-		rel->rd_istemp ? MyBackendId : InvalidBackendId;
+		rel->rd_backend == MyBackendId ? MyBackendId : InvalidBackendId;
 #else
 	reader->smgr.smgr_rnode = rel->rd_node;
 #endif
@@ -873,7 +875,7 @@ heap_is_visible(Relation heapRel, ItemPointer htid)
 	InitDirtySnapshot(SnapshotDirty);
 
 	/*
-	 * Visiblilty checking is simplifed comapred with _bt_check_unique
+	 * Visibility checking is simplified compared with _bt_check_unique
 	 * because we have exclusive lock on the relation. (XXX: Is it true?)
 	 */
 	return heap_hot_search(htid, heapRel, &SnapshotDirty, NULL);
