@@ -10,7 +10,12 @@
  */
 #include "pg_bulkload.h"
 
+#include <unistd.h>
+#include <time.h>
+
 #include "access/heapam.h"
+#include "access/reloptions.h"
+#include "commands/dbcommands.h"
 #include "commands/tablecmds.h"
 #include "funcapi.h"
 #include "miscadmin.h"
@@ -27,7 +32,9 @@
 #include "reader.h"
 #include "writer.h"
 #include "pg_btree.h"
+#include "pg_loadstatus.h"
 #include "pg_profile.h"
+#include "pg_strutil.h"
 #include "pgut/pgut-be.h"
 
 PG_MODULE_MAGIC;
@@ -37,6 +44,7 @@ PG_FUNCTION_INFO_V1(pg_bulkload);
 Datum	PGUT_EXPORT pg_bulkload(PG_FUNCTION_ARGS);
 
 static char *timeval_to_cstring(struct timeval tp);
+static void ParseOptions(Datum options, Reader **rd, Writer **wt, time_t tm);
 
 #ifdef ENABLE_BULKLOAD_PROFILE
 static instr_time prof_init;
@@ -187,8 +195,11 @@ pg_bulkload(PG_FUNCTION_ARGS)
 	 * STEP 1: Initialization
 	 */
 
-	/* parse options and create reader */
-	rd = ReaderCreate(options, ru0.tv.tv_sec);
+	/* parse options and create reader and writer */
+	ParseOptions(options, &rd, &wt, ru0.tv.tv_sec);
+
+	/* initialize reader */
+	ReaderInit(rd);
 
 	/*
 	 * We need to split PG_TRY block because gcc optimizes if-branches with
@@ -197,26 +208,26 @@ pg_bulkload(PG_FUNCTION_ARGS)
 	 */
 	PG_TRY();
 	{
-		/* truncate heap if not parallel writer */
-		if (rd->wo.truncate && rd->writer != CreateParallelWriter)
-			TruncateTable(rd->relid);
+		/* truncate heap */
+		if (wt->truncate)
+			TruncateTable(wt->relid);
 
-		/* open relation without any relation locks */
-		rd->rel = heap_open(rd->relid, NoLock);
-
-		/* initialize parser */
-		ParserInit(rd->parser, &rd->checker, rd->infile, RelationGetDescr(rd->rel));
+		/* initialize writer */
+		WriterInit(wt);
 
 		/* initialize checker */
-		CheckerInit(&rd->checker, rd->rel);
+		CheckerInit(&rd->checker, wt->rel, wt->tchecker);
 
-		/* create writer */
-		wt = rd->writer(rd->relid, &rd->wo);
+		/* initialize parser */
+		ParserInit(rd->parser, &rd->checker, rd->infile, wt->desc,
+				   wt->multi_process);
 	}
 	PG_CATCH();
 	{
 		if (rd)
 			ReaderClose(rd, true);
+		if (wt)
+			WriterClose(wt, true);
 		PG_RE_THROW();
 	}
 	PG_END_TRY();
@@ -226,7 +237,7 @@ pg_bulkload(PG_FUNCTION_ARGS)
 	PG_TRY();
 	{
 		/* create logger */
-		CreateLogger(rd->logfile, rd->wo.verbose, rd->infile[0] == ':');
+		CreateLogger(rd->logfile, wt->verbose, rd->infile[0] == ':');
 
 		start = timeval_to_cstring(ru0.tv);
 		LoggerLog(INFO, "\npg_bulkload %s on %s\n\n",
@@ -234,6 +245,7 @@ pg_bulkload(PG_FUNCTION_ARGS)
 
 		ReaderDumpParams(rd);
 		WriterDumpParams(wt);
+		LoggerLog(INFO, "\n");
 
 		BULKLOAD_PROFILE(&prof_init);
 
@@ -433,3 +445,210 @@ timeval_to_cstring(struct timeval tp)
 				TimestampTzGetDatum(tz)));
 	return str;
 }
+
+static void
+ParseOptions(Datum options, Reader **rd, Writer **wt, time_t tm)
+{
+	List		   *defs;
+	List		   *rest_defs = NIL;
+	ListCell	   *cell;
+	DefElem		   *opt;
+	char		   *keyword;
+	char		   *value;
+	char		   *type = NULL;
+	char		   *writer = NULL;
+	bool			multi_process = false;
+
+	Assert(*rd == NULL);
+	Assert(*wt == NULL);
+
+	/* parse for each option */
+	defs = untransformRelOptions(options);
+	foreach (cell, defs)
+	{
+		opt = lfirst(cell);
+		if (opt->arg == NULL)
+			ereport(ERROR, (errcode(ERRCODE_INVALID_PARAMETER_VALUE),
+					errmsg("option \"%s\" has no value", opt->defname)));
+
+		keyword = opt->defname;
+		value = strVal(opt->arg);
+
+		if (CompareKeyword(keyword, "TYPE"))
+		{
+			ASSERT_ONCE(type == NULL);
+			type = value;
+		}
+		else if (CompareKeyword(keyword, "WRITER") ||
+				 CompareKeyword(keyword, "LOADER"))
+		{
+			ASSERT_ONCE(writer == NULL);
+			writer = value;
+		}
+		else if (CompareKeyword(keyword, "MULTI_PROCESS"))
+		{
+			multi_process = ParseBoolean(value);
+		}
+		else
+		{
+			rest_defs = lappend(rest_defs, opt);
+			continue;
+		}
+	}
+
+	*wt = WriterCreate(writer, multi_process);
+	*rd = ReaderCreate(type);
+
+	foreach (cell, rest_defs)
+	{
+		opt = lfirst(cell);
+		keyword = opt->defname;
+		value = strVal(opt->arg);
+
+		if (!WriterParam((*wt), keyword, value) &&
+			!ReaderParam((*rd), keyword, value))
+		{
+			ereport(ERROR,
+				(errcode(ERRCODE_UNDEFINED_OBJECT),
+				 errmsg("invalid keyword \"%s\"", keyword)));
+		}
+	}
+
+	/*
+	 * checking necessary common setting items
+	 */
+	if ((*rd)->infile == NULL)
+		ereport(ERROR, (errcode(ERRCODE_INVALID_PARAMETER_VALUE),
+						errmsg("INPUT option required")));
+	if ((*wt)->output == NULL)
+		ereport(ERROR, (errcode(ERRCODE_INVALID_PARAMETER_VALUE),
+						errmsg("OUTPUT option required")));
+
+	/*
+	 * If log paths are not specified, generate them with the following rules:
+	 *   {basename}        : $PGDATA/pg_bulkload/YYYYMMDDHHMISS_{db}_{nsp}_{tbl}
+	 *   LOGFILE           : {basename}.log
+	 *   PARSE_BADFILE     : {basename}.prs.{extension-of-infile}
+	 *   DUPLICATE_BADFILE : {basename}.dup.csv
+	 */
+	if ((*rd)->logfile == NULL ||
+		(*rd)->parse_badfile == NULL ||
+		(*wt)->dup_badfile == NULL)
+	{
+		char		path[MAXPGPATH] = BULKLOAD_LSF_DIR "/";
+		int			len;
+		int			elen;
+		char	   *dbname;
+
+		if (getcwd(path, MAXPGPATH) == NULL)
+			elog(ERROR, "could not get current working directory: %m");
+		len = strlen(path);
+		len += snprintf(path + len, MAXPGPATH - len, "/%s/", BULKLOAD_LSF_DIR);
+		len += strftime(path + len, MAXPGPATH - len, "%Y%m%d%H%M%S_",
+						localtime(&tm));
+
+		/* TODO: cleanup characters in object names, ex. ?, *, etc. */
+		dbname = get_database_name(MyDatabaseId);
+		if ((*wt)->relid != InvalidOid)
+		{
+			char   *nspname;
+			char   *relname;
+
+			nspname = get_namespace_name(get_rel_namespace((*wt)->relid));
+			relname = get_rel_name((*wt)->relid);
+			len += snprintf(path + len, MAXPGPATH - len, "%s_%s_%s.",
+						dbname, nspname, relname);
+			pfree(nspname);
+			pfree(relname);
+		}
+		pfree(dbname);
+
+		if (len >= MAXPGPATH)
+			ereport(ERROR,
+					(errcode(ERRCODE_INTERNAL_ERROR),
+					 errmsg("default loader output file name is too long")));
+
+		canonicalize_path(path);
+
+		if ((*rd)->logfile == NULL)
+		{
+			elen = snprintf(path + len, MAXPGPATH - len, "log");
+			if (elen + len >= MAXPGPATH)
+				ereport(ERROR,
+						(errcode(ERRCODE_INTERNAL_ERROR),
+						 errmsg("default loader log file name is too long")));
+
+			(*rd)->logfile = pstrdup(path);
+		}
+
+		if ((*rd)->parse_badfile == NULL)
+		{
+			char   *filename;
+			char   *extension;
+
+			/* find the extension of infile */
+			filename = strrchr((*rd)->infile, '/');
+			extension = strrchr((*rd)->infile, '.');
+			if (!filename || !extension || filename >= extension)
+				extension = "";
+
+			elen = snprintf(path + len, MAXPGPATH - len, "prs%s", extension);
+			if (elen + len >= MAXPGPATH)
+				ereport(ERROR,
+						(errcode(ERRCODE_INTERNAL_ERROR),
+						 errmsg("default parse bad file name is too long")));
+
+			(*rd)->parse_badfile = pstrdup(path);
+		}
+
+		if ((*wt)->dup_badfile == NULL)
+		{
+			elen = snprintf(path + len, MAXPGPATH - len, "dup.csv");
+			if (elen + len >= MAXPGPATH)
+				ereport(ERROR,
+						(errcode(ERRCODE_INTERNAL_ERROR),
+						 errmsg("default duplicate bad file name is too long")));
+
+			(*wt)->dup_badfile = pstrdup(path);
+		}
+
+		/* Verify DataDir/pg_bulkload directory */
+		ValidateLSFDirectory(BULKLOAD_LSF_DIR);
+	}
+
+	/*
+	 * check it whether there is not the same file name.
+	 */
+	if ((*wt)->relid != InvalidOid &&
+		(strcmp((*rd)->infile, (*rd)->logfile) == 0 ||
+		 strcmp((*rd)->infile, (*rd)->parse_badfile) == 0 ||
+		 strcmp((*rd)->infile, (*wt)->dup_badfile) == 0 ||
+		 strcmp((*rd)->logfile, (*rd)->parse_badfile) == 0 ||
+		 strcmp((*rd)->logfile, (*wt)->dup_badfile) == 0 ||
+		 strcmp((*rd)->parse_badfile, (*wt)->dup_badfile) == 0))
+		ereport(ERROR,
+			(errcode(ERRCODE_INVALID_PARAMETER_VALUE),
+			 errmsg("INPUT, PARSE_BADFILE, DUPLICATE_BADFILE and LOGFILE cannot set the same file name.")));
+
+	if ((*wt)->relid == InvalidOid)
+	{
+		char	ctlpath[MAXPGPATH];
+
+		snprintf(ctlpath, MAXPGPATH, "%s.ctl", (*wt)->output);
+
+		if (strcmp((*rd)->infile, (*rd)->logfile) == 0 ||
+			strcmp((*rd)->infile, (*rd)->parse_badfile) == 0 ||
+			strcmp((*rd)->infile, (*wt)->output) == 0 ||
+			strcmp((*rd)->infile, ctlpath) == 0 ||
+			strcmp((*rd)->logfile, (*rd)->parse_badfile) == 0 ||
+			strcmp((*rd)->logfile, (*wt)->output) == 0 ||
+			strcmp((*rd)->logfile, ctlpath) == 0 ||
+			strcmp((*rd)->parse_badfile, (*wt)->output) == 0 ||
+			strcmp((*rd)->parse_badfile, ctlpath) == 0)
+		ereport(ERROR,
+			(errcode(ERRCODE_INVALID_PARAMETER_VALUE),
+			 errmsg("INPUT, PARSE_BADFILE, OUTPUT, LOGFILE and sample control file cannot set the same file name.")));
+	}
+
+	(*wt)->logfile = pstrdup((*rd)->logfile);
+}
