@@ -11,21 +11,13 @@
 #include "pg_bulkload.h"
 
 #include <fcntl.h>
-#include <time.h>
-#include <unistd.h>
 
 #include "access/heapam.h"
-#include "access/reloptions.h"
-#include "catalog/catalog.h"
-#include "catalog/namespace.h"
 #include "catalog/pg_proc.h"
 #include "catalog/pg_type.h"
-#include "commands/dbcommands.h"
-#include "executor/executor.h"
-#include "executor/spi.h"
 #include "funcapi.h"
 #include "mb/pg_wchar.h"
-#include "miscadmin.h"
+#include "parser/parse_coerce.h"
 #include "pgstat.h"
 #include "utils/builtins.h"
 #include "utils/lsyscache.h"
@@ -33,200 +25,73 @@
 #include "utils/resowner.h"
 #include "utils/syscache.h"
 #include "utils/typcache.h"
-#include "storage/fd.h"
-#include "tcop/tcopprot.h"
 
 #include "logger.h"
-#include "pg_loadstatus.h"
 #include "pg_profile.h"
 #include "pg_strutil.h"
-#include "reader.h"
-#include "writer.h"
 #include "pgut/pgut-be.h"
+#include "reader.h"
 
 #define DEFAULT_MAX_PARSE_ERRORS		0
-#define DEFAULT_MAX_DUP_ERRORS			0
-
-/*
- * prototype declaration of internal function
- */
-static void ParseOption(Reader *rd, DefElem *opt);
 
 /**
- * @brief Initialize Reader
- *
- * Processing flow
- *	 -# open relation
- *	 -# open input file
- *	 -# open data file
- *	 -# get function information for type information and type transformation.
- *
- * @param fname   [in] path of the control file (absolute path)
- * @param options [in] additonal options
- * @return reader.
+ * @brief Create Reader
  */
 Reader *
-ReaderCreate(Datum options, time_t tm)
+ReaderCreate(char *type)
 {
+	const char *keys[] =
+	{
+		"BINARY",
+		"FIXED",	/* alias for backward compatibility. */
+		"CSV",
+		"TUPLE",
+		"FUNCTION",
+	};
+	const ParserCreate values[] =
+	{
+		CreateBinaryParser,
+		CreateBinaryParser,
+		CreateCSVParser,
+		CreateTupleParser,
+		CreateFunctionParser,
+	};
+
 	Reader	   *self;
-	List	   *defs;
-	ListCell   *cell;
+
+	/* default of type is CSV */
+	if (type == NULL)
+		type = "CSV";
 
 	self = palloc0(sizeof(Reader));
 	self->max_parse_errors = -2;
-	self->wo.max_dup_errors = -2;
 	self->limit = INT64_MAX;
 	self->checker.encoding = -1;
 
-	/* parse for each option */
-	defs = untransformRelOptions(options);
-	foreach (cell, defs)
-		ParseOption(self, lfirst(cell));
+	self->parser = values[choice("TYPE", type, keys, lengthof(keys))]();
 
-	/*
-	 * checking necessary common setting items
-	 */
-	if (self->parser == NULL)
-		ereport(ERROR, (errcode(ERRCODE_INVALID_PARAMETER_VALUE),
-						errmsg("TYPE option required")));
-	if (self->relid == InvalidOid)
-		ereport(ERROR, (errcode(ERRCODE_INVALID_PARAMETER_VALUE),
-						errmsg("TABLE option required")));
-	if (self->infile == NULL)
-		ereport(ERROR, (errcode(ERRCODE_INVALID_PARAMETER_VALUE),
-						errmsg("INFILE option required")));
+	return self;
+}
 
+void
+ReaderInit(Reader *self)
+{
 	/*
 	 * Set defaults to unspecified parameters.
 	 */
-	if (self->writer == NULL)
-		self->writer = CreateDirectWriter;
 	if (self->max_parse_errors < -1)
 		self->max_parse_errors = DEFAULT_MAX_PARSE_ERRORS;
-	if (self->wo.max_dup_errors < -1)
-		self->wo.max_dup_errors = DEFAULT_MAX_DUP_ERRORS;
 
 	/*
-	 * Use the client_encoding case of ENCODING is not specified and INFILE is
+	 * Use the client_encoding case of ENCODING is not specified and INPUT is
 	 * STDIN.
 	 */
 	if (self->checker.encoding == -1 &&
 		pg_strcasecmp(self->infile, "stdin") == 0)
 		self->checker.encoding = pg_get_client_encoding();
-
-	/*
-	 * If log paths are not specified, generate them with the following rules:
-	 *   {basename}        : $PGDATA/pg_bulkload/YYYYMMDDHHMISS_{db}_{nsp}_{tbl}
-	 *   LOGFILE           : {basename}.log
-	 *   PARSE_BADFILE     : {basename}.prs.{extension-of-infile}
-	 *   DUPLICATE_BADFILE : {basename}.dup.csv
-	 */
-	if (self->logfile == NULL ||
-		self->parse_badfile == NULL ||
-		self->wo.dup_badfile == NULL)
-	{
-		char		path[MAXPGPATH] = BULKLOAD_LSF_DIR "/";
-		int			len;
-		int			elen;
-		char	   *dbname;
-		char	   *nspname;
-		char	   *relname;
-
-		if (getcwd(path, MAXPGPATH) == NULL)
-			elog(ERROR, "could not get current working directory: %m");
-		len = strlen(path);
-		len += snprintf(path + len, MAXPGPATH - len, "/%s/", BULKLOAD_LSF_DIR);
-		len += strftime(path + len, MAXPGPATH - len, "%Y%m%d%H%M%S_",
-						localtime(&tm));
-
-		/* TODO: cleanup characters in object names, ex. ?, *, etc. */
-		dbname = get_database_name(MyDatabaseId);
-		nspname = get_namespace_name(get_rel_namespace(self->relid));
-		relname = get_rel_name(self->relid);
-		len += snprintf(path + len, MAXPGPATH - len, "%s_%s_%s.",
-						dbname, nspname, relname);
-		pfree(dbname);
-		pfree(nspname);
-		pfree(relname);
-
-		if (len >= MAXPGPATH)
-			ereport(ERROR,
-					(errcode(ERRCODE_INTERNAL_ERROR),
-					 errmsg("default loader output file name is too long")));
-
-		canonicalize_path(path);
-
-		if (self->logfile == NULL)
-		{
-			elen = snprintf(path + len, MAXPGPATH - len, "log");
-			if (elen + len >= MAXPGPATH)
-				ereport(ERROR,
-						(errcode(ERRCODE_INTERNAL_ERROR),
-						 errmsg("default loader log file name is too long")));
-
-			self->logfile = pstrdup(path);
-		}
-
-		if (self->parse_badfile == NULL)
-		{
-			char   *filename;
-			char   *extension;
-
-			/* find the extension of infile */
-			filename = strrchr(self->infile, '/');
-			extension = strrchr(self->infile, '.');
-			if (!filename || !extension || filename >= extension)
-				extension = "";
-
-			elen = snprintf(path + len, MAXPGPATH - len, "prs%s", extension);
-			if (elen + len >= MAXPGPATH)
-				ereport(ERROR,
-						(errcode(ERRCODE_INTERNAL_ERROR),
-						 errmsg("default parse bad file name is too long")));
-
-			self->parse_badfile = pstrdup(path);
-		}
-
-		if (self->wo.dup_badfile == NULL)
-		{
-			elen = snprintf(path + len, MAXPGPATH - len, "dup.csv");
-			if (elen + len >= MAXPGPATH)
-				ereport(ERROR,
-						(errcode(ERRCODE_INTERNAL_ERROR),
-						 errmsg("default duplicate bad file name is too long")));
-
-			self->wo.dup_badfile = pstrdup(path);
-		}
-
-		/* Verify DataDir/pg_bulkload directory */
-		ValidateLSFDirectory(BULKLOAD_LSF_DIR);
-	}
-
-	/*
-	 * check it whether there is not the same file name.
-	 */
-	if (strcmp(self->infile, self->logfile) == 0 ||
-		strcmp(self->infile, self->parse_badfile) == 0 ||
-		strcmp(self->infile, self->wo.dup_badfile) == 0 ||
-		strcmp(self->logfile, self->parse_badfile) == 0 ||
-		strcmp(self->logfile, self->wo.dup_badfile) == 0 ||
-		strcmp(self->parse_badfile, self->wo.dup_badfile) == 0)
-		ereport(ERROR,
-			(errcode(ERRCODE_INVALID_PARAMETER_VALUE),
-			 errmsg("INFILE, PARSE_BADFILE, DUPLICATE_BADFILE and LOGFILE cannot set the same file name.")));
-
-	self->wo.logfile = self->logfile;
-
-	return self;
 }
 
-const char *ON_DUPLICATE_NAMES[] =
-{
-	"NEW",
-	"OLD"
-};
-
-static size_t
+size_t
 choice(const char *name, const char *key, const char *keys[], size_t nkeys)
 {
 	size_t		i;
@@ -244,34 +109,15 @@ choice(const char *name, const char *key, const char *keys[], size_t nkeys)
 
 /**
  * @brief Parse a line in control file.
- * @param rd   [in] reader
- * @param opt  [in] option
- * @return None
  */
-static void
-ParseOption(Reader *rd, DefElem *opt)
+bool
+ReaderParam(Reader *rd, const char *keyword, char *target)
 {
-	char	   *keyword;
-	char	   *target;
-
-	if (opt->arg == NULL)
-		ereport(ERROR, (errcode(ERRCODE_INVALID_PARAMETER_VALUE),
-				errmsg("option \"%s\" has no value", opt->defname)));
-
-	keyword = opt->defname;
-	target = strVal(opt->arg);
-
 	/*
 	 * result
 	 */
-	if (CompareKeyword(keyword, "TABLE"))
-	{
-		ASSERT_ONCE(rd->relid == InvalidOid);
-
-		rd->relid = RangeVarGetRelid(makeRangeVarFromNameList(
-						stringToQualifiedNameList(target)), false);
-	}
-	else if (CompareKeyword(keyword, "INFILE"))
+	if (CompareKeyword(keyword, "INFILE") ||
+		CompareKeyword(keyword, "INPUT"))
 	{
 		ASSERT_ONCE(rd->infile == NULL);
 
@@ -289,53 +135,6 @@ ParseOption(Reader *rd, DefElem *opt)
 
 		rd->parse_badfile = pstrdup(target);
 	}
-	else if (CompareKeyword(keyword, "DUPLICATE_BADFILE"))
-	{
-		ASSERT_ONCE(rd->wo.dup_badfile == NULL);
-
-		rd->wo.dup_badfile = pstrdup(target);
-	}
-	else if (CompareKeyword(keyword, "TYPE"))
-	{
-		const char *keys[] =
-		{
-			"BINARY",
-			"FIXED",	/* alias for backward compatibility. */
-			"CSV",
-			"TUPLE",
-			"FUNCTION",
-		};
-		const ParserCreate values[] =
-		{
-			CreateBinaryParser,
-			CreateBinaryParser,
-			CreateCSVParser,
-			CreateTupleParser,
-			CreateFunctionParser,
-		};
-
-		ASSERT_ONCE(rd->parser == NULL);
-		rd->parser = values[choice(keyword, target, keys, lengthof(keys))]();
-	}
-	else if (CompareKeyword(keyword, "WRITER") ||
-			 CompareKeyword(keyword, "LOADER"))
-	{
-		const char *keys[] =
-		{
-			"DIRECT",
-			"BUFFERED",
-			"PARALLEL"
-		};
-		const WriterCreate values[] =
-		{
-			CreateDirectWriter,
-			CreateBufferedWriter,
-			CreateParallelWriter
-		};
-
-		ASSERT_ONCE(rd->writer == NULL);
-		rd->writer = values[choice(keyword, target, keys, lengthof(keys))];
-	}
 	else if (CompareKeyword(keyword, "PARSE_ERRORS") ||
 			 CompareKeyword(keyword, "MAX_ERR_CNT"))
 	{
@@ -344,37 +143,12 @@ ParseOption(Reader *rd, DefElem *opt)
 		if (rd->max_parse_errors == -1)
 			rd->max_parse_errors = INT64_MAX;
 	}
-	else if (CompareKeyword(keyword, "DUPLICATE_ERRORS"))
-	{
-		ASSERT_ONCE(rd->wo.max_dup_errors < -1);
-		rd->wo.max_dup_errors = ParseInt64(target, -1);
-		if (rd->wo.max_dup_errors == -1)
-			rd->wo.max_dup_errors = INT64_MAX;
-	}
 	else if (CompareKeyword(keyword, "LOAD") ||
 			 CompareKeyword(keyword, "LIMIT"))
 	{
 		ASSERT_ONCE(rd->limit == INT64_MAX);
 		rd->limit = ParseInt64(target, 0);
 	}
-	else if (CompareKeyword(keyword, "ON_DUPLICATE_KEEP"))
-	{
-		const ON_DUPLICATE values[] =
-		{
-			ON_DUPLICATE_KEEP_NEW,
-			ON_DUPLICATE_KEEP_OLD
-		};
-
-		rd->wo.on_duplicate = values[choice(keyword, target, ON_DUPLICATE_NAMES, lengthof(values))];
-	}
-	else if (CompareKeyword(keyword, "VERBOSE"))
-	{
-		rd->wo.verbose = ParseBoolean(target);
-	}
-	else if (CompareKeyword(keyword, "TRUNCATE"))
-	{
-		rd->wo.truncate = ParseBoolean(target);
-	}
 	else if (CompareKeyword(keyword, "CHECK_CONSTRAINTS"))
 	{
 		rd->checker.check_constraints = ParseBoolean(target);
@@ -391,11 +165,9 @@ ParseOption(Reader *rd, DefElem *opt)
 	}
 	else if (rd->parser == NULL ||
 			!ParserParam(rd->parser, keyword, target))
-	{
-		ereport(ERROR,
-			(errcode(ERRCODE_UNDEFINED_OBJECT),
-			 errmsg("invalid keyword \"%s\"", keyword)));
-	}
+		return false;
+
+	return true;
 }
 
 /**
@@ -415,9 +187,6 @@ ReaderClose(Reader *rd, bool onError)
 
 	CheckerTerm(&rd->checker);
 
-	if (rd->rel)
-		heap_close(rd->rel, NoLock);
-
 	if (!onError)
 	{
 		if (rd->parse_fp != NULL && FreeFile(rd->parse_fp) < 0)
@@ -431,8 +200,6 @@ ReaderClose(Reader *rd, bool onError)
 			pfree(rd->logfile);
 		if (rd->parse_badfile != NULL)
 			pfree(rd->parse_badfile);
-		if (rd->wo.dup_badfile != NULL)
-			pfree(rd->wo.dup_badfile);
 
 		pfree(rd);
 	}
@@ -467,7 +234,11 @@ ReaderNext(Reader *rd)
 			if (tuple == NULL)
 				eof = true;
 			else
+			{
+				tuple = CheckerTuple(&rd->checker, tuple,
+									 &parser->parsing_field);
 				CheckerConstraints(&rd->checker, tuple, &parser->parsing_field);
+			}
 		}
 		PG_CATCH();
 		{
@@ -546,41 +317,23 @@ ReaderNext(Reader *rd)
 void
 ReaderDumpParams(Reader *self)
 {
-	char		   *tablename;
-	char		   *nspname;
-	char		   *relname;
 	char		   *str;
 	StringInfoData	buf;
 
 	initStringInfo(&buf);
 
 	str = QuoteString(self->infile);
-	appendStringInfo(&buf, "INFILE = %s\n", str);
+	appendStringInfo(&buf, "INPUT = %s\n", str);
 	pfree(str);
 
 	str = QuoteString(self->parse_badfile);
 	appendStringInfo(&buf, "PARSE_BADFILE = %s\n", str);
 	pfree(str);
 
-	str = QuoteString(self->wo.dup_badfile);
-	appendStringInfo(&buf, "DUPLICATE_BADFILE = %s\n", str);
-	pfree(str);
-
 	str = QuoteString(self->logfile);
 	appendStringInfo(&buf, "LOGFILE = %s\n", str);
 	pfree(str);
 
-	nspname = get_namespace_name(get_rel_namespace(self->relid));
-	relname = get_rel_name(self->relid);
-	tablename = quote_qualified_identifier(nspname, relname);
-	str = QuoteString(tablename);
-	appendStringInfo(&buf, "TABLE = %s\n", tablename);
-	pfree(str);
-	pfree(tablename);
-	pfree(nspname);
-	pfree(relname);
-
-	appendStringInfo(&buf, "TRUNCATE = %s\n", self->wo.truncate ? "YES" : "NO");
 	if (self->limit == INT64_MAX)
 		appendStringInfo(&buf, "LIMIT = INFINITE\n");
 	else
@@ -591,14 +344,6 @@ ReaderDumpParams(Reader *self)
 	else
 		appendStringInfo(&buf, "PARSE_ERRORS = " int64_FMT "\n",
 						 self->max_parse_errors);
-	if (self->wo.max_dup_errors == INT64_MAX)
-		appendStringInfo(&buf, "DUPLICATE_ERRORS = INFINITE\n");
-	else
-		appendStringInfo(&buf, "DUPLICATE_ERRORS = " int64_FMT "\n",
-						 self->wo.max_dup_errors);
-	appendStringInfo(&buf, "ON_DUPLICATE_KEEP = %s\n",
-					 ON_DUPLICATE_NAMES[self->wo.on_duplicate]);
-	appendStringInfo(&buf, "VERBOSE = %s\n", self->wo.verbose ? "YES" : "NO");
 	if (PG_VALID_FE_ENCODING(self->checker.encoding))
 		appendStringInfo(&buf, "ENCODING = %s\n",
 						 pg_encoding_to_char(self->checker.encoding));
@@ -612,10 +357,12 @@ ReaderDumpParams(Reader *self)
 }
 
 void
-CheckerInit(Checker *checker, Relation rel)
+CheckerInit(Checker *checker, Relation rel, TupleChecker *tchecker)
 {
 	TupleDesc	desc;
 
+	checker->tchecker = tchecker;
+
 	/*
 	 * When specify ENCODING, we check the input data encoding.
 	 * Convert encoding if the client and server encodings are different.
@@ -626,6 +373,9 @@ CheckerInit(Checker *checker, Relation rel)
 		checker->db_encoding != PG_SQL_ASCII))
 		checker->check_encoding = true;
 
+	if (!rel)
+		return;
+
 	/* When specify CHECK_CONSTRAINTS, we check the constraints */
 	desc = RelationGetDescr(rel);
 	if (desc->constr &&
@@ -732,7 +482,7 @@ CheckerConversion(Checker *checker, char *src)
 											  checker->db_encoding);
 }
 
-void
+HeapTuple
 CheckerConstraints(Checker *checker, HeapTuple tuple, int *parsing_field)
 {
 	if (checker->has_constraints)
@@ -766,6 +516,8 @@ CheckerConstraints(Checker *checker, HeapTuple tuple, int *parsing_field)
 			}
 		}
 	}
+
+	return tuple;
 }
 
 void
@@ -910,7 +662,7 @@ TupleFormerValue(TupleFormer *former, const char *str, int col)
  * destination type, so long as the physical storage matches.  This is
  * helpful in some cases involving out-of-date cached plans.
  */
-void
+bool
 tupledesc_match(TupleDesc dst_tupdesc, TupleDesc src_tupdesc)
 {
 	int			i;
@@ -937,34 +689,27 @@ tupledesc_match(TupleDesc dst_tupdesc, TupleDesc src_tupdesc)
 		if (dattr->atttypid == sattr->atttypid)
 			continue;			/* no worries */
 		if (!dattr->attisdropped)
-			ereport(ERROR,
-					(errcode(ERRCODE_DATATYPE_MISMATCH),
-					 errmsg("function return row and target table row do not match"),
-					 errdetail("Returned type %s at ordinal position %d, but target table %s.",
-							   format_type_be(sattr->atttypid),
-							   i + 1,
-							   format_type_be(dattr->atttypid))));
+			return false;
 
 		if (dattr->attlen != sattr->attlen ||
 			dattr->attalign != sattr->attalign)
-			ereport(ERROR,
-					(errcode(ERRCODE_DATATYPE_MISMATCH),
-					 errmsg("function return row and target table row do not match"),
-					 errdetail("Physical storage mismatch on dropped attribute at ordinal position %d.",
-							   i + 1)));
+			return false;
 	}
+
+	return true;
 }
 
-void
+TupleCheckStatus
 FilterInit(Filter *filter, TupleDesc desc)
 {
 	int				i;
 	ParsedFunction	func;
 	HeapTuple		ftup;
 	Form_pg_proc	pp;
+	TupleCheckStatus	status = NEED_COERCION_CHECK;
 
 	if (filter->funcstr == NULL)
-		return;
+		return NO_COERCION;
 
 	/* parse filter function */
 	func = ParseFunction(filter->funcstr, true);
@@ -993,21 +738,8 @@ FilterInit(Filter *filter, TupleDesc desc)
 				 errmsg("filter function must not return set")));
 
 	/* Check data type of the function result value */
-	if (pp->prorettype == desc->tdtypeid)
-	{
-		/*
-		 * If the return value of the filter function is a target table,
-		 * lookup_rowtype_tupdesc grab AccessShareLock on the table in the
-		 * first call.  We call lookup_rowtype_tupdesc here to avoid deadlock
-		 * when lookup_rowtype_tupdesc is called by the internal routine of the
-		 * filter function, because a parallel writer process holds an
-		 * AccessExclusiveLock.
-		 */
-		TupleDesc	resultDesc = lookup_rowtype_tupdesc(pp->prorettype, -1);
-		ReleaseTupleDesc(resultDesc);
-
-		filter->tupledesc_matched = true;
-	}
+	if (pp->prorettype == desc->tdtypeid && pp->prorettype != RECORDOID)
+		status = NO_COERCION;
 	else if (pp->prorettype == RECORDOID)
 	{
 		TupleDesc	resultDesc = NULL;
@@ -1017,8 +749,9 @@ FilterInit(Filter *filter, TupleDesc desc)
 
 		if (resultDesc)
 		{
-			tupledesc_match(desc, resultDesc);
-			filter->tupledesc_matched = true;
+			if (tupledesc_match(desc, resultDesc))
+				status = NO_COERCION;
+
 			FreeTupleDesc(resultDesc);
 		}
 	}
@@ -1087,6 +820,8 @@ FilterInit(Filter *filter, TupleDesc desc)
 	filter->fn_rettype = pp->prorettype;
 
 	ReleaseSysCache(ftup);
+
+	return status;
 }
 
 void
@@ -1186,51 +921,166 @@ FilterTuple(Filter *filter, TupleFormer *former, int *parsing_field)
 	if (fcinfo.isnull)
 		return TupleFormerNullTuple(former);
 
-	/* Check tupdesc of the function result value */
-	if (!filter->tupledesc_matched)
-	{
-		HeapTupleHeader	td = DatumGetHeapTupleHeader(datum);
-
-		/*
-		 * We must not call lookup_rowtype_tupdesc, because parallel writer
-		 * process a deadlock could occur, when typeid same target table's row
-		 * type.
-		 */
-		if (former->desc->tdtypeid != HeapTupleHeaderGetTypeId(td))
-		{
-			TupleDesc		resultDesc;
-
-			resultDesc = lookup_rowtype_tupdesc(HeapTupleHeaderGetTypeId(td),
-												HeapTupleHeaderGetTypMod(td));
-#ifdef ALL_RESULTDESC_CHECK
-			PG_TRY();
-			{
-#endif
-				tupledesc_match(former->desc, resultDesc);
-#ifdef ALL_RESULTDESC_CHECK
-			}
-			PG_CATCH();
-			{
-				ReleaseTupleDesc(resultDesc);
-				if (filter->fn_rettype == RECORDOID)
-					*parsing_field = 0;
-
-				PG_RE_THROW();
-			}
-			PG_END_TRY();
-
-			if (filter->fn_rettype != RECORDOID)
-#endif
-				filter->tupledesc_matched = true;
-
-			ReleaseTupleDesc(resultDesc);
-		}
-		else
-			filter->tupledesc_matched = true;
-	}
-
 	filter->tuple.t_data = DatumGetHeapTupleHeader(datum);
 	filter->tuple.t_len = HeapTupleHeaderGetDatumLength(filter->tuple.t_data);
 
 	return &filter->tuple;
 }
+
+TupleChecker *
+CreateTupleChecker(TupleDesc desc)
+{
+	TupleChecker   *self;
+	MemoryContext	context;
+	MemoryContext	oldcontext;
+
+	context = AllocSetContextCreate(CurrentMemoryContext,
+									"TupleChecker",
+									ALLOCSET_DEFAULT_MINSIZE,
+									ALLOCSET_DEFAULT_INITSIZE,
+									ALLOCSET_DEFAULT_MAXSIZE);
+
+	oldcontext = MemoryContextSwitchTo(context);
+
+	self = palloc0(sizeof(TupleChecker));
+	self->status = NEED_COERCION_CHECK;
+	self->sourceDesc = NULL;
+	self->targetDesc = CreateTupleDescCopy(desc);
+	self->context = context;
+	self->values = (Datum *) palloc(desc->natts * sizeof(Datum));
+	self->nulls = (bool *) palloc(desc->natts * sizeof(bool));
+	self->opt = NULL;
+	self->coercionChecker = NULL;
+
+	MemoryContextSwitchTo(oldcontext);
+
+	return self;
+}
+
+void
+UpdateTupleCheckStatus(TupleChecker *self, HeapTuple tuple)
+{
+	HeapTupleHeader		td;
+	Oid					typeid;
+	TupleDesc			resultDesc;
+	MemoryContext		oldcontext;
+
+	td = tuple->t_data;
+	typeid = HeapTupleHeaderGetTypeId(td);
+
+	/*
+	 * We must not call lookup_rowtype_tupdesc, because parallel writer
+	 * process a deadlock could occur, when typeid same target table's row
+	 * type.
+	 */
+	if (self->targetDesc->tdtypeid == typeid && typeid != RECORDOID)
+	{
+		self->status = NO_COERCION;
+		return;
+	}
+
+	resultDesc = lookup_rowtype_tupdesc(typeid,
+										HeapTupleHeaderGetTypMod(td));
+	if (tupledesc_match(self->targetDesc, resultDesc))
+	{
+		self->status = NO_COERCION;
+		ReleaseTupleDesc(resultDesc);
+		return;
+	}
+
+	self->status = NEED_COERCION;
+
+	oldcontext = MemoryContextSwitchTo(self->context);
+	self->sourceDesc = CreateTupleDescCopy(resultDesc);
+	MemoryContextSwitchTo(oldcontext);
+	ReleaseTupleDesc(resultDesc);
+}
+
+void
+CoercionDeformTuple(TupleChecker *self, HeapTuple tuple, int *parsing_field)
+{
+	int	i;
+	int	natts;
+
+	natts = self->targetDesc->natts;
+
+	if (self->typIsVarlena == NULL)
+	{
+		Oid				iofunc;
+		MemoryContext	oldcontext;
+
+		oldcontext = MemoryContextSwitchTo(self->context);
+		self->typIsVarlena = (bool *) palloc(natts * sizeof(bool));
+		self->typOutput = (FmgrInfo *) palloc(natts * sizeof(FmgrInfo));
+		self->typIOParam = (Oid *) palloc(natts * sizeof(Oid));
+		self->typInput = (FmgrInfo *) palloc(natts * sizeof(FmgrInfo));
+
+		for (i = 0; i < natts; i++)
+		{
+			if (self->sourceDesc->attrs[i]->atttypid ==
+				self->targetDesc->attrs[i]->atttypid)
+				continue;
+
+			getTypeOutputInfo(self->sourceDesc->attrs[i]->atttypid,
+							  &iofunc, &self->typIsVarlena[i]);
+			fmgr_info(iofunc, &self->typOutput[i]);
+
+			getTypeInputInfo(self->targetDesc->attrs[i]->atttypid, &iofunc,
+							 &self->typIOParam[i]);
+			fmgr_info(iofunc, &self->typInput[i]);
+		}
+
+		MemoryContextSwitchTo(oldcontext);
+	}
+	
+	heap_deform_tuple(tuple, self->sourceDesc, self->values, self->nulls);
+
+	for (i = 0; i < natts; i++)
+	{
+		*parsing_field = i + 1;
+
+		/* Ignore dropped columns in datatype */
+		if (self->targetDesc->attrs[i]->attisdropped)
+			continue;
+
+		if (self->nulls[i])
+		{
+			/* emit nothing... */
+			continue;
+		}
+		else if (self->sourceDesc->attrs[i]->atttypid ==
+				 self->targetDesc->attrs[i]->atttypid)
+		{
+			continue;
+		}
+		else
+		{
+			char   *value;
+
+			value = OutputFunctionCall(&self->typOutput[i], self->values[i]);
+			self->values[i] = InputFunctionCall(&self->typInput[i], value,
+										self->typIOParam[i],
+										self->targetDesc->attrs[i]->atttypmod);
+		}
+	}
+
+	*parsing_field = -1;
+}
+
+HeapTuple
+CoercionCheckerTuple(TupleChecker *self, HeapTuple tuple, int *parsing_field)
+{
+	HeapTuple	retTuple;
+
+	if (self->status == NEED_COERCION_CHECK)
+		UpdateTupleCheckStatus(self, tuple);
+
+	if (self->status == NO_COERCION)
+		return tuple;
+
+	CoercionDeformTuple(self, tuple, parsing_field);
+
+	retTuple = heap_form_tuple(self->targetDesc, self->values, self->nulls);
+
+	return retTuple;
+}
